@@ -62,13 +62,23 @@
 
 #define CALIBRATE_COUNT 0x800   // Approx 1.7ms
 
-u32 cpu_khz VAR16VISIBLE;
-u8 no_tsc VAR16VISIBLE;
+u32 cpu_khz VARFSEG;
+u8 no_tsc VARFSEG;
+
+u16 pmtimer_ioport VARFSEG;
+u32 pmtimer_wraps VARLOW;
+u32 pmtimer_last VARLOW;
 
 static void
 calibrate_tsc(void)
 {
     u32 eax, ebx, ecx, edx, cpuid_features = 0;
+
+    if (CONFIG_PMTIMER && GET_GLOBAL(pmtimer_ioport)) {
+        dprintf(3, "pmtimer already configured; will not calibrate TSC\n");
+        return;
+    }
+
     cpuid(0, &eax, &ebx, &ecx, &edx);
     if (eax > 0)
         cpuid(1, &eax, &ebx, &ecx, &cpuid_features);
@@ -108,30 +118,59 @@ calibrate_tsc(void)
     dprintf(1, "CPU Mhz=%u\n", hz / 1000000);
 }
 
+/* TSC emulation timekeepers */
+u64 TSC_8254 VARLOW;
+int Last_TSC_8254 VARLOW;
+
 static u64
 emulate_tsc(void)
 {
-    int cnt, d;
-    u16 ebda_seg = get_ebda_seg();
-    u64 ret;
     /* read timer 0 current count */
-    ret = GET_EBDA2(ebda_seg, tsc_8254);
-    /* readback mode has slightly shifted registers, works on all 8254, readback PIT0 latch */
+    u64 ret = GET_LOW(TSC_8254);
+    /* readback mode has slightly shifted registers, works on all
+     * 8254, readback PIT0 latch */
     outb(PM_SEL_READBACK | PM_READ_VALUE | PM_READ_COUNTER0, PORT_PIT_MODE);
-    cnt = (inb(PORT_PIT_COUNTER0) | (inb(PORT_PIT_COUNTER0) << 8));
-    d = GET_EBDA2(ebda_seg, last_tsc_8254) - cnt;
+    int cnt = (inb(PORT_PIT_COUNTER0) | (inb(PORT_PIT_COUNTER0) << 8));
+    int d = GET_LOW(Last_TSC_8254) - cnt;
     /* Determine the ticks count from last invocation of this function */
     ret += (d > 0) ? d : (PIT_TICK_INTERVAL + d);
-    SET_EBDA2(ebda_seg, last_tsc_8254, cnt);
-    SET_EBDA2(ebda_seg, tsc_8254, ret);
+    SET_LOW(Last_TSC_8254, cnt);
+    SET_LOW(TSC_8254, ret);
     return ret;
 }
 
+void pmtimer_setup(u16 ioport, u32 khz)
+{
+    if (!CONFIG_PMTIMER)
+        return;
+    dprintf(1, "Using pmtimer, ioport 0x%x, freq %d kHz\n", ioport, khz);
+    SET_GLOBAL(pmtimer_ioport, ioport);
+    SET_GLOBAL(cpu_khz, khz);
+}
+
+static u64 pmtimer_get(void)
+{
+    u16 ioport = GET_GLOBAL(pmtimer_ioport);
+    u32 wraps = GET_LOW(pmtimer_wraps);
+    u32 pmtimer = inl(ioport) & 0xffffff;
+
+    if (pmtimer < GET_LOW(pmtimer_last)) {
+        wraps++;
+        SET_LOW(pmtimer_wraps, wraps);
+    }
+    SET_LOW(pmtimer_last, pmtimer);
+
+    dprintf(9, "pmtimer: %u:%u\n", wraps, pmtimer);
+    return (u64)wraps << 24 | pmtimer;
+}
+
 static u64
 get_tsc(void)
 {
     if (unlikely(GET_GLOBAL(no_tsc)))
         return emulate_tsc();
+    if (CONFIG_PMTIMER && GET_GLOBAL(pmtimer_ioport))
+        return pmtimer_get();
     return rdtscll();
 }
 
@@ -233,7 +272,7 @@ pit_setup(void)
 }
 
 static void
-init_rtc(void)
+rtc_setup(void)
 {
     outb_cmos(0x26, CMOS_STATUS_A);    // 32,768Khz src, 976.5625us updates
     u8 regB = inb_cmos(CMOS_STATUS_B);
@@ -248,6 +287,8 @@ bcd2bin(u8 val)
     return (val & 0xf) + ((val >> 4) * 10);
 }
 
+u8 Century VARLOW;
+
 void
 timer_setup(void)
 {
@@ -255,7 +296,7 @@ timer_setup(void)
     calibrate_tsc();
     pit_setup();
 
-    init_rtc();
+    rtc_setup();
     rtc_updating();
     u32 seconds = bcd2bin(inb_cmos(CMOS_RTC_SECONDS));
     u32 minutes = bcd2bin(inb_cmos(CMOS_RTC_MINUTES));
@@ -264,6 +305,18 @@ timer_setup(void)
     ticks = ((u64)ticks * PIT_TICK_RATE) / PIT_TICK_INTERVAL;
     SET_BDA(timer_counter, ticks);
 
+    // Setup Century storage
+    if (CONFIG_QEMU) {
+        Century = inb_cmos(CMOS_CENTURY);
+    } else {
+        // Infer current century from the year.
+        u8 year = inb_cmos(CMOS_RTC_YEAR);
+        if (year > 0x80)
+            Century = 0x19;
+        else
+            Century = 0x20;
+    }
+
     enable_hwirq(0, FUNC16(entry_08));
     enable_hwirq(8, FUNC16(entry_70));
 }
@@ -360,7 +413,7 @@ handle_1a03(struct bregs *regs)
     // Bit4 in try#1 flipped in hardware (forced low) due to bit7=1
     // My assumption: RegB = ((RegB & 01100000b) | 00000010b)
     if (rtc_updating()) {
-        init_rtc();
+        rtc_setup();
         // fall through as if an update were not in progress
     }
     outb_cmos(regs->dh, CMOS_RTC_SECONDS);
@@ -387,14 +440,7 @@ handle_1a04(struct bregs *regs)
     regs->cl = inb_cmos(CMOS_RTC_YEAR);
     regs->dh = inb_cmos(CMOS_RTC_MONTH);
     regs->dl = inb_cmos(CMOS_RTC_DAY_MONTH);
-    if (CONFIG_COREBOOT) {
-        if (regs->cl > 0x80)
-            regs->ch = 0x19;
-        else
-            regs->ch = 0x20;
-    } else {
-        regs->ch = inb_cmos(CMOS_CENTURY);
-    }
+    regs->ch = GET_LOW(Century);
     regs->al = regs->ch;
     set_success(regs);
 }
@@ -414,15 +460,14 @@ handle_1a05(struct bregs *regs)
     // Bit4 in try#1 flipped in hardware (forced low) due to bit7=1
     // My assumption: RegB = (RegB & 01111111b)
     if (rtc_updating()) {
-        init_rtc();
+        rtc_setup();
         set_invalid(regs);
         return;
     }
     outb_cmos(regs->cl, CMOS_RTC_YEAR);
     outb_cmos(regs->dh, CMOS_RTC_MONTH);
     outb_cmos(regs->dl, CMOS_RTC_DAY_MONTH);
-    if (!CONFIG_COREBOOT)
-        outb_cmos(regs->ch, CMOS_CENTURY);
+    SET_LOW(Century, regs->ch);
     // clear halt-clock bit
     u8 val8 = inb_cmos(CMOS_STATUS_B) & ~RTC_B_SET;
     outb_cmos(val8, CMOS_STATUS_B);
@@ -453,7 +498,7 @@ handle_1a06(struct bregs *regs)
         return;
     }
     if (rtc_updating()) {
-        init_rtc();
+        rtc_setup();
         // fall through as if an update were not in progress
     }
     outb_cmos(regs->dh, CMOS_RTC_SECONDS_ALARM);
@@ -507,7 +552,6 @@ handle_1a(struct bregs *regs)
     case 0x05: handle_1a05(regs); break;
     case 0x06: handle_1a06(regs); break;
     case 0x07: handle_1a07(regs); break;
-    case 0xb1: handle_1ab1(regs); break;
     default:   handle_1aXX(regs); break;
     }
 }
@@ -518,8 +562,7 @@ handle_08(void)
 {
     debug_isr(DEBUG_ISR_08);
 
-    floppy_tick();
-
+    // Update counter
     u32 counter = GET_BDA(timer_counter);
     counter++;
     // compare to one days worth of timer ticks at 18.2 hz
@@ -528,14 +571,17 @@ handle_08(void)
         counter = 0;
         SET_BDA(timer_rollover, GET_BDA(timer_rollover) + 1);
     }
-
     SET_BDA(timer_counter, counter);
 
+    // Check for internal events.
+    floppy_tick();
     usb_check_event();
 
     // chain to user timer tick INT #0x1c
-    u32 eax=0, flags;
-    call16_simpint(0x1c, &eax, &flags);
+    struct bregs br;
+    memset(&br, 0, sizeof(br));
+    br.flags = F_IF;
+    call16_int(0x1c, &br);
 
     eoi_pic1();
 }
@@ -545,12 +591,13 @@ handle_08(void)
  * Periodic timer
  ****************************************************************/
 
+int RTCusers VARLOW;
+
 void
 useRTC(void)
 {
-    u16 ebda_seg = get_ebda_seg();
-    int count = GET_EBDA2(ebda_seg, RTCusers);
-    SET_EBDA2(ebda_seg, RTCusers, count+1);
+    int count = GET_LOW(RTCusers);
+    SET_LOW(RTCusers, count+1);
     if (count)
         return;
     // Turn on the Periodic Interrupt timer
@@ -561,9 +608,8 @@ useRTC(void)
 void
 releaseRTC(void)
 {
-    u16 ebda_seg = get_ebda_seg();
-    int count = GET_EBDA2(ebda_seg, RTCusers);
-    SET_EBDA2(ebda_seg, RTCusers, count-1);
+    int count = GET_LOW(RTCusers);
+    SET_LOW(RTCusers, count-1);
     if (count != 1)
         return;
     // Clear the Periodic Interrupt.
@@ -610,7 +656,7 @@ handle_1586(struct bregs *regs)
         return;
     }
     while (!statusflag)
-        wait_irq();
+        yield_toirq();
     set_success(regs);
 }
 
@@ -667,8 +713,10 @@ handle_70(void)
         goto done;
     if (registerC & RTC_B_AIE) {
         // Handle Alarm Interrupt.
-        u32 eax=0, flags;
-        call16_simpint(0x4a, &eax, &flags);
+        struct bregs br;
+        memset(&br, 0, sizeof(br));
+        br.flags = F_IF;
+        call16_int(0x4a, &br);
     }
     if (!(registerC & RTC_B_PIE))
         goto done;
