@@ -9,7 +9,7 @@
 #include "util.h" // handle_1ab1
 #include "pci.h" // pci_config_readl
 #include "bregs.h" // struct bregs
-#include "biosvar.h" // GET_EBDA
+#include "biosvar.h" // GET_GLOBAL
 #include "pci_regs.h" // PCI_VENDOR_ID
 
 // romlayout.S
@@ -133,11 +133,12 @@ handle_1ab10d(struct bregs *regs)
 static void
 handle_1ab10e(struct bregs *regs)
 {
-    struct pir_header *pirtable_g = (void*)(GET_GLOBAL(PirOffset) + 0);
-    if (! pirtable_g) {
+    struct pir_header *pirtable_gf = GET_GLOBAL(PirAddr);
+    if (! pirtable_gf) {
         set_code_invalid(regs, RET_FUNC_NOT_SUPPORTED);
         return;
     }
+    struct pir_header *pirtable_g = GLOBALFLAT2GLOBAL(pirtable_gf);
 
     struct param_s {
         u16 size;
@@ -200,19 +201,19 @@ handle_1ab1(struct bregs *regs)
     }
 }
 
+// Entry point for pci bios functions.
+void VISIBLE16 VISIBLE32SEG
+handle_pcibios(struct bregs *regs)
+{
+    debug_enter(regs, DEBUG_HDL_pcibios);
+    handle_1ab1(regs);
+}
+
 
 /****************************************************************
  * 32bit interface
  ****************************************************************/
 
-// Entry point for 32bit pci bios functions.
-void VISIBLE32SEG
-handle_pcibios32(struct bregs *regs)
-{
-    debug_enter(regs, DEBUG_HDL_pcibios32);
-    handle_1ab1(regs);
-}
-
 struct bios32_s {
     u32 signature;
     u32 entry;
@@ -222,13 +223,13 @@ struct bios32_s {
     u8 reserved[5];
 } PACKED;
 
-struct bios32_s BIOS32HEADER __aligned(16) VAR16EXPORT = {
+struct bios32_s BIOS32HEADER __aligned(16) VARFSEG = {
     .signature = 0x5f32335f, // _32_
     .length = sizeof(BIOS32HEADER) / 16,
 };
 
 void
-bios32_setup(void)
+bios32_init(void)
 {
     dprintf(3, "init bios32\n");
 
